<?php

/**
 * Copyright (c) 2020 CodeLathe. All rights Reserved.
 * This file is part of FileCloud  https://www.filecloud.com
 */

define('TONIDO_CLOUD_ROOT_DIR', realpath(__DIR__ . '/../../..'));
require_once(TONIDO_CLOUD_ROOT_DIR . DIRECTORY_SEPARATOR . 'thirdparty' . DIRECTORY_SEPARATOR . 'vendor' . DIRECTORY_SEPARATOR . 'autoload.php');
$cloudconfigphp = TONIDO_CLOUD_ROOT_DIR . DIRECTORY_SEPARATOR . 'config' . DIRECTORY_SEPARATOR . 'cloudconfig.php';
require_once($cloudconfigphp);

define('MONGO_DEFAULT_PORT', 27017);

class MongoUtility
{
    public string $uri;
    public array $uriOptions = [];
    public array $driverOptions = [];

    public array $dbnodes = [];

    public MongoDB\Driver\Manager $manager;

    protected string $primaryMember = '';
    protected string $replicaSetStatus = '';

    public string $mongoExe;
    public string $mongoDumpExe;
    public string $mongoParameters = '';

    public function console(string $msg): void
    {
        $logline = date("Y/m/d h:i:s") . " $msg" . PHP_EOL;
        echo $logline;
    }

    /**
     * @param $msg
     * @return void
     */
    public function error($msg): void
    {
        $logline = date("Y/m/d h:i:s") . " Error: $msg";
        echo "\033[01;31m" . $logline . " \033[0m" . PHP_EOL;
    }

    function checkFileAccess()
    {
        // check if files are readable
        if (isset($this->uriOptions['tlsCertificateKeyFile'])) {
            $fileName = $this->uriOptions['tlsCertificateKeyFile'];
            if (!is_readable($fileName)) {
                $this->error('file not readable: \'' . $fileName . '\'!');
            }
        }
        if (isset($uriOptions['tlsCAFile'])) {
            $fileName = $this->uriOptions['tlsCAFile'];
            if (!is_readable($fileName)) {
                $this->error('file not readable: \'' . $fileName . '\'!');
            }
        }
    }

    private function setMongoBin()
    {
        if (strtoupper(substr(PHP_OS, 0, 3)) !== 'WIN') {
            // mongo bin folder is included in PATH, no absolute path needed for linux
            $this->mongoExe = 'mongo';
            $this->mongoDumpExe = 'mongodump';
        } else {
            $this->mongoExe = TONIDO_CLOUD_ROOT_DIR . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . 'mongodb' . DIRECTORY_SEPARATOR . 'bin' . DIRECTORY_SEPARATOR . 'mongo.exe';
            $this->mongoDumpExe = TONIDO_CLOUD_ROOT_DIR . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . 'mongodb' . DIRECTORY_SEPARATOR . 'bin' . DIRECTORY_SEPARATOR . 'mongodump.exe';
        }
    }

    private function setMongoParameters()
    {
        if (function_exists('FC_MONGODB_URI_OPTIONS')) {
            $this->uriOptions = call_user_func('FC_MONGODB_URI_OPTIONS');
        }

        if (function_exists('FC_MONGODB_DRIVER_OPTIONS')) {
            $this->driverOptions = call_user_func('FC_MONGODB_DRIVER_OPTIONS');
        }

        $this->uri = '';
        if (defined('TONIDOCLOUD_DBSERVER')) {
            $this->uri = constant('TONIDOCLOUD_DBSERVER');
        }

        if (preg_match('/mongodb:\/\/(.*)@/', $this->uri, $matches)) {
            $credentials = $matches[1];
            $parts = explode(':', $credentials);
            $this->uriOptions['username'] = $parts[0];
            $this->uriOptions['password'] = $parts[1];
            $this->uri = preg_replace("/mongodb:\/\/.*@/", "mongodb://", $this->uri);
        }

        $uriparts = explode('/', $this->uri, 4);
        $nodes = explode(',', $uriparts[2]);
        foreach ($nodes as $node) {
            $nodeparts = explode(':', $node,);
            $nodeName=$nodeparts[0];
            $this->dbnodes[$nodeName]['nodeName'] = $nodeName;
            $this->dbnodes[$nodeName]['uri'] = $uriparts[0] . '//' . $node . '/' . $uriparts[3];
            if(isset($nodeparts[1])){
                $this->dbnodes[$nodeName]['port'] = $uriparts[1];
            } else {
                $this->dbnodes[$nodeName]['port'] = MONGO_DEFAULT_PORT;
            }
        }

        if (!empty($this->uriOptions)) {
            foreach ($this->uriOptions as $name => $value) {
                if ($value == 1) {
                    $this->mongoParameters .= ' --' . $name;
                } else {
                    $this->mongoParameters .= ' --' . $name . '=' . $value;
                }
            }
        }
    }

    public function __construct()
    {
        if (php_sapi_name() == 'cli') {
            error_reporting(E_ALL & ~E_NOTICE);
            ini_set("log_errors", 1);
            // redirect php errors to stdout
            ini_set('error_log', '');
            ini_set('max_execution_time', '60');
        }
        $this->setMongoParameters();
        $this->checkFileAccess();
        $this->setMongoBin();

        $this->manager = new MongoDB\Driver\Manager($this->uri, $this->uriOptions);
    }

    public function getRsStatus()
    {
        if (str_contains(TONIDOCLOUD_DBSERVER, 'replicaSet')) {
            $command = new MongoDB\Driver\Command(['replSetGetStatus' => 1]);
            $cursor = $this->manager->executeCommand('admin', $command);
            $response = current($cursor->toArray());
            $clusterState = "ok";
            $this->console("");
            $this->console(">>> replica set -----------------------------------------------------");
            if ($response->ok != 1) {
                $this->error("failed to read replica set status!");
            } else {
                // process response
                foreach ($response->members as $member) {
                    $this->console("member $member->_id, name: $member->name, state: $member->stateStr, last heartbeat message: '$member->lastHeartbeatMessage'");
                    if ($member->stateStr == 'PRIMARY') {
                        $this->primaryMember = $member->name;
                    }
                    if (!$member->health) {
                        $clusterState = "ERROR";
                    }
                }
            }
            $this->replicaSetStatus = $clusterState;
            $this->console("cluster state: $clusterState");
            $this->console("<<< replica set -----------------------------------------------------");
            $this->console("");
        } else {
            $this->console("No replica set configured.");
        }
    }

    public function getServerStatus()
    {

        $this->console("");
        $this->console(">>> server info -----------------------------------------------------");

        foreach ($this->dbnodes as $node ) {
            $manager = new MongoDB\Driver\Manager($node['uri'], $this->uriOptions);
            $command = new MongoDB\Driver\Command(['serverStatus' => 1]);
            $t1 = microtime(true);
            $cursor = $manager->executeCommand('admin', $command);
            $t2 = microtime(true);
            $diff = round(($t2 - $t1), 3) * 1000;
            $response = current($cursor->toArray());
            if (!$response->ok) {
                $this->error("failed to read server status for '" .  $node['nodeName'] .  "!");
            } else {
                $this->console("member: " . $node['nodeName']  . ", host: $response->host, version: $response->version, time: $diff ms.");
            }
        }
        $this->console("<<< server info -----------------------------------------------------");
        $this->console("");
    }

    public
    function getFeatureCompatibilityVersion()
    {
        $command = new MongoDB\Driver\Command(['getParameter' => 1, 'featureCompatibilityVersion' => 1]);
        $cursor = $this->manager->executeCommand('admin', $command);
        $response = current($cursor->toArray());
        if ($response->ok != 1) {
            $this->error("failed to get featureCompatibilityVersion !");
        } else {
            $version = $response->featureCompatibilityVersion->version;
            $this->console("featureCompatibilityVersion: $version");
        }
    }

    public
    function setFeatureCompatibilityVersion(string $version)
    {
        $command = new MongoDB\Driver\Command(['setFeatureCompatibilityVersion' => $version]);
        $cursor = $this->manager->executeCommand('admin', $command);
        $response = current($cursor->toArray());
        if ($response->ok != 1) {
            $this->error("failed to set featureCompatibilityVersion !");
        } else {
            $this->console("featureCompatibilityVersion was set to $version.");
        }
    }

    public function rsStepDown()
    {
        $this->getRsStatus();
        $oldPrimary = $this->primaryMember;
        if ($this->replicaSetStatus != 'ok') {
            $this->error("replicaset has status '$this->replicaSetStatus', cannot do a stepDown!");
        } else {
            $command = new MongoDB\Driver\Command(['replSetStepDown' => '60']);
            $cursor = $this->manager->executeCommand('admin', $command);
            $response = current($cursor->toArray());
            if ($response->ok != 1) {
                $this->error("failed to start stepDown !");
            } else {
                $this->console("initiated stepDown for member $this->primaryMember");
            }
            $sleepTimeSeconds = 10;
            $this->console("waiting $sleepTimeSeconds seconds for the stepDown to finish ...");
            sleep($sleepTimeSeconds);
            $this->getRsStatus();
            $newPrimary = $this->primaryMember;
            if ($oldPrimary != $newPrimary && $this->replicaSetStatus == 'ok') {
                $this->console("stepDown for member $this->primaryMember finished, new PRIMARY: $this->primaryMember.");
            } else {
                $this->error("stepDown for member $this->primaryMember failed!");
            }
        }
    }

    /**
     * @return array
     * @throws \MongoDB\Driver\Exception\Exception
     */
    public function getFcDatabases(){
        $command = new MongoDB\Driver\Command(['listDatabases' => 1]);
        $cursor = $this->manager->executeCommand('admin', $command);
        $response = current($cursor->toArray());
        $fcDatabases = [];
        foreach($response->databases as $database) {
            $fcDatabases[] = (string) $database->name;
        }
        return $fcDatabases;
    }
}
